unit uMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  WSocket, ExtCtrls, StdCtrls, ComCtrls;

const
   WM_REMOVECLI = WM_USER + 1;
   WM_CONNECT   = WM_USER + 2;
   BUF_SIZE     = 255;           // receiving buffer size w/o terminating null
   DATA_TX      = '123456789.123456789.123456789.123456789.123456789.' +
                  '123456789.123456789.123456789.123456789.123456789.' +
                  '123456789.123456789.123456789.123456789.123456789.' +
                  '123456789.123456789.123456789.123456789.123456789.' +
                  '123456789.123456789.123456789.123456789.123456789.' +
                  '123456789.123456789.123456789.123456789.123456789.' +
                  '123456789.123456789.123456789.123456789.123456789.' +
                  '123456789.123456789.123456789.123456789.123456789.' +
                  '123456789.123456789.123456789.123456789.123456789.' +
                  '123456789.123456789.123456789.123456789.123456789.';

type
  TGprsSocket = class(TWSocket)
  public
    Sending: boolean;
    LoggedOn: boolean;
    LogonID: string;
    MsgIdMob: integer;
    TimeToSend: cardinal;
  end;

type
  TMain = class(TForm)
    Panel1: TPanel;
    Activate: TCheckBox;
    Panel2: TPanel;
    GroupBox1: TGroupBox;
    CountEdt: TEdit;
    ClientCount: TUpDown;
    CliCount: TStaticText;
    GroupBox2: TGroupBox;
    Label1: TLabel;
    Label2: TLabel;
    Addr: TEdit;
    Port: TEdit;
    Panel3: TPanel;
    Memo: TMemo;
    ID: TEdit;
    Label3: TLabel;
    DataSizeEdt: TEdit;
    Label4: TLabel;
    DataSize: TUpDown;
    Label5: TLabel;
    IntervalEdt: TEdit;
    Interval: TUpDown;
    IntervalTimer: TTimer;
    Label6: TLabel;
    BackLogDlyEdt: TEdit;
    BackLogDly: TUpDown;
    ClearBtn: TButton;
    GroupBox3: TGroupBox;
    Label7: TLabel;
    Label8: TLabel;
    PacketsSent: TStaticText;
    PacketsAck: TStaticText;
    ConnectTimer: TTimer;
    procedure ActivateClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure IntervalTimerTimer(Sender: TObject);
    procedure ClearBtnClick(Sender: TObject);
    procedure ConnectTimerTimer(Sender: TObject);
  private
    SentCount: integer;
    AckCount: integer;
    LogonNo: integer;
    HostIP: string;
    ClientList: TList;
    procedure StartClients;
    procedure StopClients;
    procedure CliDataAvailable(Sender: TObject; Error: Word);
    procedure CliDnsLookupDone(Sender: TObject; Error: Word);
    procedure CliSessionClosed(Sender: TObject; Error: Word);
    procedure CliSessionConnected(Sender: TObject; Error: Word);
    procedure WMRemoveCli(var Msg: TMessage); message WM_REMOVECLI;
    procedure WMConnect(var Msg: TMessage); message WM_CONNECT;
    procedure Log(const Data: string);
    procedure Ack(Cli: TObject; Data: PChar);
    procedure Rcvd(Cli: TObject; Data: PChar);
  public
  end;

var
  Main: TMain;

implementation

{$R *.DFM}

//------------------------------------------------------------------------------
procedure TMain.WMRemoveCli(var Msg: TMessage);
var
   Cli: TWSocket;
begin
   try
      Cli := TWSocket(Msg.WParam);
      Cli.Destroy;
      ClientList.Remove(Cli);
      CliCount.Caption := IntToStr(ClientList.Count);
      if ClientList.Count = 0 then
         Activate.Checked := False;
   except
   end;
end;

//------------------------------------------------------------------------------
procedure TMain.Log(const Data: string);
begin
   Memo.Lines.Add(Data);
end;

//------------------------------------------------------------------------------
procedure TMain.FormCreate(Sender: TObject);
begin
   ClientList := TList.Create;
end;

//------------------------------------------------------------------------------
procedure TMain.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
   StopClients;
   ClientList.Free;
end;

//------------------------------------------------------------------------------
procedure TMain.WMConnect(var Msg: TMessage);
var
   Cli: TGPRSSocket;
begin
   if not Activate.Checked then
      Exit;

   try
      Cli := TGPRSSocket.Create(nil);
      Cli.Addr    := HostIP;
      Cli.Port    := Port.Text;
      Cli.LogonID := ID.Text + '_' + IntToStr(LogonNo);
      Cli.OnDataAvailable    := CliDataAvailable;
      Cli.OnSessionClosed    := CliSessionClosed;
      Cli.OnSessionConnected := CliSessionConnected;
      Cli.Connect;
      ClientList.Add(Cli);
      CliCount.Caption := IntToStr(ClientList.Count);
      Inc(LogonNo);
   except
      on E: Exception do begin
         Log(E.Message);
         Activate.Checked := False;
      end;
   end;
end;

//------------------------------------------------------------------------------
procedure TMain.StartClients;
var
   Cli: TWSocket;
begin
   SentCount := 0;
   AckCount  := 0;
   PacketsSent.Caption := '0';
   PacketsAck.Caption  := '0';
   Cli := TWSocket.Create(Self);
   Cli.OnDnsLookupDone := CliDnsLookupDone;
   Cli.DnsLookup(Addr.Text);
end;

//------------------------------------------------------------------------------
procedure TMain.StopClients;
var
   n: integer;
begin
   IntervalTimer.Enabled := False;
   for n := 0 to ClientList.Count - 1 do
      TWSocket(ClientList[n]).CloseDelayed;
end;

//------------------------------------------------------------------------------
procedure TMain.ClearBtnClick(Sender: TObject);
begin
   Memo.Clear;
end;

//------------------------------------------------------------------------------
procedure TMain.ActivateClick(Sender: TObject);
begin
   with Sender as TCheckBox do
   try
      CountEdt.Enabled      := not Checked;
      ClientCount.Enabled   := not Checked;
      Addr.Enabled          := not Checked;
      Port.Enabled          := not Checked;
      DataSize.Enabled      := not Checked;
      DataSizeEdt.Enabled   := not Checked;
      Interval.Enabled      := not Checked;
      IntervalEdt.Enabled   := not Checked;
      ID.Enabled            := not Checked;
      BackLogDly.Enabled    := not Checked;
      BackLogDlyEdt.Enabled := not Checked;
      if Checked then
         StartClients
      else
         StopClients;
   except
      on E: Exception do begin
         Checked := False;
         Log(E.Message);
      end;
   end;
end;

//------------------------------------------------------------------------------
procedure TMain.CliDnsLookupDone(Sender: TObject; Error: Word);
begin
   try
      if Error <> 0 then begin
         Log(IntToStr(Error) + ' ' + WSocketErrorDesc(Error));
         Exit;
      end;
      if not Activate.Checked then
         Exit;

      HostIP := TWSocket(Sender).DnsResult;
      PostMessage(Handle, WM_CONNECT, 0, 0);
      ConnectTimer.Interval := BacklogDly.Position;
      ConnectTimer.Enabled := True;
      IntervalTimer.Enabled  := True;
   finally
      TWSocket(Sender).Release;
   end;
end;

//------------------------------------------------------------------------------
procedure TMain.CliSessionClosed(Sender: TObject; Error: Word);
begin
   if Error <> 0 then
      Log(IntToStr(Error) + ' ' + WSocketErrorDesc(Error));
   TGprsSocket(Sender).LoggedOn := False;
   PostMessage(Handle, WM_REMOVECLI, Integer(Sender), 0);
   if Activate.Checked and not ConnectTimer.Enabled then
      ConnectTimer.Enabled := True;
end;

//------------------------------------------------------------------------------
procedure TMain.CliSessionConnected(Sender: TObject; Error: Word);
begin
   if Error <> 0 then begin
      Log(IntToStr(Error) + ' ' + WSocketErrorDesc(Error));
      Exit;
   end;
   try
      with Sender as TGPRSSocket do begin
         SendStr('L:' + LogonID + #13#10);
         Sending := True;
      end;
   except
      on E: Exception do
         Log(E.Message);
   end;
end;

//------------------------------------------------------------------------------
procedure TMain.CliDataAvailable(Sender: TObject; Error: Word);
var
   Buf: array[0.. BUF_SIZE] of char;
   Count: integer;
begin
   if Error <> 0 then begin
      Log(IntToStr(Error) + ' ' + WSocketErrorDesc(Error));
      Exit;
   end;
   Count := TWSocket(Sender).Receive(@Buf[0], BUF_SIZE - 1);
   if Count <= 2 then
      Exit;
   Buf[Count - 2] := #0;
   case Buf[0] of
      'A': Ack(Sender, Buf);
      'D': Rcvd(Sender, Buf);
   end;
end;

//------------------------------------------------------------------------------
procedure TMain.Ack(Cli: TObject; Data: PChar);
begin
   TGprsSocket(Cli).Sending := False;
   if Data[1] = 'L' then begin
      TGprsSocket(Cli).LoggedOn := True;
      TGprsSocket(Cli).TimeToSend := GetTickCount;
   end
   else begin
      Inc(AckCount);
      PacketsAck.Caption := IntToStr(AckCount);
      PacketsAck.Update;
   end;
end;

//------------------------------------------------------------------------------
procedure TMain.Rcvd(Cli: TObject; Data: PChar);
begin
   Log(Data);
end;

//------------------------------------------------------------------------------
procedure TMain.IntervalTimerTimer(Sender: TObject);
var
   n: integer;
   Cli: TGprsSocket;
begin
   for n := 0 to ClientList.Count - 1 do
   try
      Cli := TGprsSocket(ClientList[n]);
      if not Cli.LoggedOn then begin
         Continue;
      end;
      if Cli.Sending then begin
         Continue;
      end;
      if Cli.TimeToSend = 0 then
         Continue;
      if GetTickCount < Cli.TimeToSend then
         Continue;

      Cli.PutStringInSendBuffer('D:' + IntToStr(Cli.MsgIdMob) + 'd');
      Cli.PutDataInSendBuffer(@DATA_TX[1], DataSize.Position);
      Cli.SendStr(#13#10);
      Inc(Cli.MsgIdMob);
      Cli.Sending := True;
      Cli.TimeToSend := GetTickCount + cardinal(Interval.Position) * 1000;
      Inc(SentCount);
      PacketsSent.Caption := IntToStr(SentCount);
      PacketsSent.Update;
   except
      on E: Exception do
         Log(E.Message);
   end;
end;

//------------------------------------------------------------------------------
procedure TMain.ConnectTimerTimer(Sender: TObject);
begin
   if ClientList.Count < ClientCount.Position then
      PostMessage(Handle, WM_CONNECT, 0, 0)
   else
      TTimer(Sender).Enabled := False;
end;

end.
